# $Id: 8a05a4ac5aa26503d0b05d4e53db7d355ab95f47 $

"""
The *Grizzled Utility Library* is a general-purpose Python library with
a variety of different modules and packages. It's roughly organized into
subpackages that group different kinds of utility functions and classes.

See the `API documentation`_ for complete details.

.. _API documentation: http://www.clapper.org/software/python/grizzled/epydoc
"""

__version__ = '0.9.4'
__author__  = 'Brian M. Clapper'
__email__   = 'bmc@clapper.org'
__url__     = 'http://bmc.github.com/grizzled/'
__license__ = 'BSD license'
__title__   = 'The Grizzled Python Utility Library'

# Public stuff

version = __version__
author  = __author__
email   = __email__
url     = __url__
license = __license__
title   = __title__
